# This is required for running on EL6
import __main__
__main__.__requires__ = ['SQLAlchemy >= 0.7', 'Flask >= 0.9', 'jinja2 >= 2.6']
import pkg_resources

import sys
import os
from optparse import OptionParser

from blockerbugs import db, app
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.release import Release
from blockerbugs.util.bug_sync import BugSync
from blockerbugs.util.update_sync import UpdateSync
from blockerbugs.util.bz_interface import BlockerBugs
from blockerbugs.models.bug import Bug
from alembic.config import Config
from alembic import command as al_command
from alembic.migration import MigrationContext


def get_alembic_config():
    # the location of the alembic ini file and alembic scripts changes when
    # installed via package
    if os.path.exists("./alembic.ini"):
        alembic_cfg = Config("./alembic.ini")
    else:
        alembic_cfg = Config("/usr/share/blockerbugs/alembic.ini",
                             ini_section='alembic-packaged')
    return alembic_cfg


def initialize_db():
    alembic_cfg = get_alembic_config()

    # check to see if the db has already been initialized by checking for an
    # alembic revision
    context = MigrationContext.configure(db.engine.connect())
    current_rev = context.get_current_revision()

    if current_rev:
        print "Database already initialized and at rev %s - not re-initializing" % current_rev
    else:
        print "Initializing Database"
        db.drop_all()
        db.create_all()

        print "Initializing alembic version"

        al_command.stamp(alembic_cfg, "head")


def upgrade_db():
    print "Upgrading Database to Latest Revision"
    alembic_cfg = get_alembic_config()
    al_command.upgrade(alembic_cfg, "head")


def add_release(number):
    existing_release = Release.query.filter_by(number=number).count()
    if existing_release > 0:
        print "Release %s already exists, not adding" % number
    else:
        print "Adding release %s" % number
        release_num = int(number)
        new_release = Release(release_num)
        db.session.add(new_release)
        db.session.commit()


def add_milestone(number, version, blocker, fe):
    existing_milestone = Milestone.query.filter_by(version=version,
                                                   blocker_tracker=blocker,
                                                   fe_tracker=fe).count()

    current_milestones = Milestone.query.filter_by(active=True).count()
    if existing_milestone > 0:
        print "Milestone already exists, not adding: %s-%s (%s, %s)" % (number, version, blocker, fe)
    else:
        print "Adding milestone: %s-%s (%s, %s)" % (number, version, blocker, fe)
        release_num = int(number)
        release = Release.query.filter_by(number=release_num).first()
        if not release:
            add_release(release)
            release = Release.query.filter_by(number=release_num).first()

        new_milestone = Milestone(release, version, int(blocker), int(fe),
                                  '%d-%s' % (release_num, version))

        # only active releases are synced but since this version is being added,
        # it's a good assumption to make it active.
        new_milestone.active = True

        # there should be only one current version, the new version as current if
        # there are no current releases defined
        if current_milestones <= 0:
            new_milestone.current = True
        db.session.add(new_milestone)
        db.session.commit()


def generate_config(dburi=None, bugzilla_cookie=None):
    config_filename = os.path.abspath('conf/settings.py')
    if os.path.exists(config_filename):
        print "configuration file %s already exists, exiting" % config_filename
        sys.exit(0)

    secret_key = os.urandom(24).encode('string_escape')

    config_file = open(config_filename, 'w')
    config_file.write('SECRET_KEY = \'%s\'\n' % secret_key)
    if dburi:
        config_file.write('SQLALCHEMY_DATABASE_URI = ')
        config_file.write('\'%s\'\n' % dburi)
    else:
        config_file.write('SQLALCHEMY_DATABASE_URI = \'\'\n')
    # leave places for the other config values generally needed
    config_file.write('FAS_ADMIN_GROUP = \'\'\n')
    config_file.write('FAS_USER = \'\'\n')
    config_file.write('FAS_PASSWORD = \'\'\n')
    if bugzilla_cookie:
        config_file.write('BUGZILLA_COOKIE = \'%s\'\n' % str(bugzilla_cookie))
    else:
        config_file.write('BUGZILLA_COOKIE = None\n')
    config_file.close()


def check_blockers(bzurl):
    bzinterface = BlockerBugs(url=bzurl)
    active_milestones = Milestone.query.filter_by(active=True).all()
    for milestone in active_milestones:
        missing_bugs = []
        buglist = bzinterface.do_query(milestone.blocker_tracker)
        for bug in buglist:
            bugid = bug.bug_id
            new_bug = Bug.query.filter_by(bugid=bugid).first()
            if not new_bug:
                missing_bugs.append(bugid)

        print "========================================"
        print "Fedora %d %s Missing Blocker Bugs" % (milestone.release.number, milestone.version)
        print "========================================"
        if len(missing_bugs) > 0:
            for bug in missing_bugs:
                print "  * %d" % bug
        else:
            print "  No missing bugs found"


def sync_bugs(fullsync=False, docheck=False):
    # grab the URL from app config so that it's syncing from the right place
    bzurl = app.config['BUGZILLA_URL']

    current_milestone = Milestone.query.filter_by(active=True).first()
    if not current_milestone:
        sys.stderr.write('BugSync ERROR: No active releases found!')
        sys.exit(1)

    sync = BugSync(db, url=bzurl)

    sync.update_active(full_sync=fullsync)

    if docheck:
        print "Checking bug sync status"
        check_blockers(bzurl)


def sync_updates(fullsync=False):
    active_releases = Release.query.filter_by(active=True).all()
    update_sync = UpdateSync(db)
    for release in active_releases:
        update_sync.sync_updates(release)


def main():
    possible_commands = ['init_db', 'add_milestone', 'add_release',
                         'generate_config', 'sync-updates', 'sync-bugs',
                         'sync', 'upgrade_db']

    usage = 'usage: %prog [options] command \n  Possible Commands: ' + ' '.join(
        possible_commands)
    parser = OptionParser(usage=usage)
    parser.add_option('-f', '--full', action='store_true', default=False,
                      help='Force full sync')
    parser.add_option('-c', '--check', action='store_true', default=False,
                      help='Force check for missing blocker bugs after sync')
    parser.add_option('-d', '--dburi', dest='dburi', help='dburi to use')
    parser.add_option('-r', '--release', dest='release', help='release number')
    parser.add_option('-m', '--milestone', dest='milestone',
                      help='Milestone name')
    parser.add_option('-b', '--blocker', dest='blocker',
                      help='blocker tracking bug id')
    parser.add_option('-a', '--accepted', dest='fe', help='FE tracking bug id')
    parser.add_option('-z', '--bzcookie', dest='bzcookie', default=None,
                      help='location of bugzilla cookie file to use')

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print "need to have at least 1 command"
        sys.exit(1)

    command = args[0]
    if not command in possible_commands:
        print 'Invalid command: %s' % command
        print 'Please use one of the following commands: %s' % str(possible_commands)
        sys.exit(1)

    if command == 'sync':
        sync_bugs(fullsync=options.full, docheck=options.check)
        sync_updates(fullsync=options.full)
    elif command == 'sync-bugs':
        sync_bugs(fullsync=options.full, docheck=options.check)
    elif command == 'sync-updates':
        sync_updates(fullsync=options.full)
    elif command == 'upgrade_db':
        upgrade_db()
    elif command == 'init_db':
        initialize_db()
    elif command == 'add_milestone':
        add_milestone(options.release, options.milestone, options.blocker,
                      options.fe)
    elif command == 'add_release':
        add_release(options.release)
    elif command == 'generate_config':
        generate_config(dburi=options.dburi, bugzilla_cookie=options.bzcookie)


if __name__ == '__main__':
    exit = main()
    if exit:
        sys.exit(exit)

