#
# spin.py - Spin view for admin interface
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
import re
from datetime import datetime

import koji
from flask import request, url_for, flash, redirect
from flask.ext import wtf
from flask.ext.admin import expose
from flask.ext.admin.babel import gettext
from flask.ext.admin.contrib.sqlamodel.filters import BaseSQLAFilter, FilterConverter
from flask.ext.admin.contrib.sqlamodel.form import AdminModelConverter
from flask.ext.admin.form import Select2Field
from flask.ext.admin.model import filters
from flask.ext.admin.model.form import converts
from wtforms import widgets

from blockerbugs.controllers.admin.auth import FasAuthModelView
from blockerbugs.models.build import Build
from blockerbugs.models.spin import Spin, SpinState, SpinType


class DeclEnumEqualFilter(BaseSQLAFilter):
    def __init__(self, column, name, options=None, data_type=None):
        super(DeclEnumEqualFilter, self).__init__(column, name, options,
                                                  data_type)

    def get_options(self, view):
        return [(k, k) for k in self.options.keys()]

    def apply(self, query, raw_value):
        value = self.options[raw_value]
        return query.filter(self.column == value)

    def operation(self):
        return gettext('equal')


class SpinFilterConvert(FilterConverter):
    _column_types = {
        'state': SpinState,
        'spin type': SpinType,
    }

    @filters.convert('DeclEnumType')
    def conv_enum(self, column, name, options=None, **kwargs):
        if not options:
            decl_enum = self._column_types[name.lower().strip()]
            options = dict((unicode(k), v) for k, v in decl_enum._reg.items())
        name = name.strip()  # flask-admin 1.0.5 does not strip column name
        return [DeclEnumEqualFilter(column, name, options, **kwargs)]


class Select2BuildsWidget(widgets.TextInput):
    def __call__(self, field, **kwargs):
        kwargs['data-role'] = u'select2Builds'
        choices = field.iter_choices()
        kwargs['existing-builds'] = u','.join(choices)
        return super(Select2BuildsWidget, self).__call__(field, **kwargs)


class SelectBuildsTagsField(wtf.Field):
    widget = Select2BuildsWidget()

    def __init__(self, *args, **kwargs):
        self._form = kwargs.get('_form', None)
        super(SelectBuildsTagsField, self).__init__(*args, **kwargs)

    def pre_validate(self, form):
        exist_builds = set(b.nvr for b in Build.query.filter(
            Build.nvr.in_(self._form_data)).all())
        for build in self._form_data:
            if build not in exist_builds:
                try:
                    Build.from_nvr(build)
                except koji.GenericError, ex:
                    self.errors.append('%s: %s' % (build, ex.message))

    def _get_data(self):
        if self._form_data is not None:
            builds = Build.query.filter(Build.nvr.in_(self._form_data)).all()
            self._set_data(builds)
        return self._data

    def _set_data(self, data):
        self._data = data
        self._form_data = None

    data = property(_get_data, _set_data)

    def process_formdata(self, valuelist):
        if valuelist:
            self._data = None
            self._form_data = list(
                v.strip() for v in valuelist[0].split(',') if v.strip())

    def iter_choices(self):
        if self._form and self._form.succeeds.data:
            builds = self._form.succeeds.data.builds
            return [b.nvr for b in builds]
        return []

    def _value(self):
        if self.raw_data:
            return u','.join(self.raw_data)
        else:
            return u','.join((b.nvr for b in self.data)) if self.data else u''


class DeclEnumModelConverter(AdminModelConverter):
    @converts('DeclEnumType')
    def conv_DeclEnum(self, column, field_args, **extra):
        field_args['validators'].append(
            wtf.validators.AnyOf(column.type.enums))
        field_args['choices'] = [(f, f) for f in column.type.enums]
        return Select2Field(**field_args)


class SpinView(FasAuthModelView):
    form_excluded_columns = ('date_created', 'date_requested')
    form_columns = ('name', 'state', 'spin_type', 'url', 'succeeds',
                    'milestone', 'updates', 'builds')
    column_searchable_list = [Spin.name]
    model_form_converter = DeclEnumModelConverter
    column_filters = (Spin.state, Spin.spin_type, )
    filter_converter = SpinFilterConvert()
    form_overrides = {'builds': SelectBuildsTagsField}
    edit_template = 'admin/spin_edit.html'
    create_template = 'admin/spin_create.html'


    _number_parse_re = re.compile(r'(.*?)(\d+)(/|)$')

    def _fill_new_spin_form(self, form):
        last_spin = Spin.query.order_by(Spin.date_requested.desc()).first()
        if last_spin is None:
            return
        form.succeeds.data = last_spin
        if last_spin.spin_type:
            form.spin_type.data = last_spin.spin_type.description
        form.milestone.data = last_spin.milestone
        form.state.data = SpinState.requested.description
        form.builds.data = last_spin.builds
        form.updates.data = last_spin.updates
        try:
            prefix, numb, _ = re.match(self._number_parse_re,
                                       last_spin.name).groups()
            assert numb.isdigit()
            form.name.data = prefix + str(int(numb) + 1)
        except AttributeError:
            form.name.data = ''
        try:
            prefix, numb, sl = re.match(self._number_parse_re,
                                        last_spin.url).groups()
            assert numb.isdigit()
            form.url.data = prefix + str(int(numb) + 1) + sl
        except AttributeError:
            form.url.data = ''


    @expose('/new/', methods=('GET', 'POST'))
    def create_view(self):
        """
            Create model view
            Fill new form with data from previous spin.
        """
        return_url = request.args.get('url') or url_for('.index_view')

        form = self.create_form()

        if form.validate_on_submit():
            if self.create_model(form):
                if '_add_another' in request.form:
                    flash(gettext('Model was successfully created.'))
                    return redirect(url_for('.create_view', url=return_url))
                else:
                    return redirect(return_url)

        if not form.is_submitted():
            self._fill_new_spin_form(form)

        return self.render(self.create_template,
                           form=form,
                           form_widget_args=self.form_widget_args,
                           return_url=return_url)
