#
# api.py - RESTful API handlers
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
#

import datetime

from flask import Blueprint, request
from iso8601 import parse_date

from blockerbugs import db
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.update import Update
from blockerbugs.models.release import Release
from blockerbugs.models.bug import Bug
from blockerbugs.models.spin import Spin, SpinState, SpinType
import validators
import errors
from utils import get_or_404, json_loads, JsonResponse, check_api_rights

api_v0 = Blueprint('api', __name__, url_prefix='/api/v0')

ACCEPTED_BUGTYPES = [
    'proposed_blocker', 'proposed_fe',
    'rejected_blocker', 'rejected_fe',
    'accepted_blocker', 'accepted_fe',
]


@api_v0.errorhandler(errors.RestApiError)
def api_error_handler(e):
    return JsonResponse(e.to_dict(), e.http_status_code)


def get_update_info(update):
    update_simple_fields = ['title', 'url', 'karma', 'stable_karma', 'status',
                            'pending']
    update_data = dict(
        (attr, getattr(update, attr)) for attr in update_simple_fields)
    update_data['release'] = update.release.number
    update_data['milestones'] = [{'version': m.version,
                                  'release': m.release.number, }
                                 for m in update.milestones]
    update_data['bugs'] = [
        {'bugid': bug.bugid,
         'type': [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]}
        for bug in update.bugs
    ]
    return update_data


def get_bug_info(bug):
    bug_simple_fields = ['bugid', 'url', 'summary', 'component', 'active']
    bug_info = dict((attr, getattr(bug, attr)) for attr in bug_simple_fields)
    bug_info['type'] = [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]
    return bug_info


def get_spin_info(spin):
    spin_simple_fields = ['id', 'name', 'url', 'date_requested',
                          'date_created', 'spin_type']
    spin_info = dict(
        (attr, getattr(spin, attr)) for attr in spin_simple_fields)
    spin_info['state'] = spin.state.value
    spin_info['updates'] = [{'title': up.title,
                             'status': up.status,
                             'url': up.url} for up in spin.updates]
    spin_info['builds'] = [{'nvr': build.nvr,
                            'epoch': build.epoch,
                            'id': build.koji_id} for build in spin.builds]
    if spin.succeeds:
        spin_info['succeeds'] = {'id': spin.succeeds.id,
                                 'name': spin.succeeds.name}
    else:
        spin_info['succeeds'] = {}
    spin_info['milestone'] = {'release': spin.milestone.release.number,
                              'version': spin.milestone.version}
    return spin_info


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/updates')
def list_updates(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            updates = updates.filter(Update.bugs.any(bug_attr == True))
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    updates = updates.order_by(Update.date_submitted.desc()).all()
    updates_info = [get_update_info(up) for up in updates]
    return JsonResponse(updates_info)


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/bugs')
def list_bugs(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    bugs = Bug.query.filter_by(milestone=milestone)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            bugs = bugs.filter(bug_attr == True)
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    bugs = bugs.order_by(Bug.component, Bug.bugid).all()
    bugs_info = [get_bug_info(bug) for bug in bugs]
    return JsonResponse(bugs_info)


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/spins')
def list_spins(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    spins = milestone.spins.order_by(Spin.date_requested.desc())
    spins_info = [get_spin_info(spin) for spin in spins]
    return JsonResponse(spins_info)


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/spins',
              methods=['POST'])
@check_api_rights
def create_spin(rel_num, milestone_version):
    spin_validator = validators.DictValidator({
        'name': validators.StringValidator(),
        'spin_type': validators.ChoicesValidator(SpinType.values()),
        'updates': validators.ListValidator(basestring),
        'succeeds': validators.IntegerValidator(required=False),
        'state': validators.ChoicesValidator(SpinState.values(),
                                             required=False),
        'url': validators.UrlValidator(required=False)
    })
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    data = json_loads(request.data)
    spin_validator(data)
    spin_kwargs = {
        'name': data['name'],
        'spin_type': data['spin_type'],
        'updates': [get_or_404(Update, title=tit) for tit in data['updates']],
        'milestone': milestone,
        'date_requested': datetime.datetime.utcnow(),
    }
    if 'succeeds' in data:
        spin_kwargs['succeeds'] = get_or_404(Spin, id=data['succeeds'])
    if 'state' in data:
        spin_kwargs['state'] = data['state']
    else:
        spin_kwargs['state'] = SpinState.requested
    if 'url' in data:
        spin_kwargs['url'] = data['url']
    spin = Spin(**spin_kwargs)
    db.session.add(spin)
    db.session.commit()
    location_str = 'milestone/{m.release.number}/{m.version}/spins/{spin.id}'
    location_header = {'Location': location_str.format(m=milestone, spin=spin)}
    return JsonResponse(status=201, headers=location_header)


@api_v0.route(
    '/milestones/<int:rel_num>/<milestone_version>/spins/<int:spin_id>')
def get_spin(rel_num, milestone_version, spin_id):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone,
                           release=release, version=milestone_version)
    spin = get_or_404(Spin, id=spin_id)
    if spin.milestone != milestone:
        raise errors.NoSuchObjectError(obj_type='Spin')
    spin_info = get_spin_info(spin)
    return JsonResponse(spin_info)


@api_v0.route(
    '/milestones/<int:rel_num>/<milestone_version>/spins/<int:spin_id>',
    methods=['PUT'])
@check_api_rights
def update_spin(rel_num, milestone_version, spin_id):
    spin_validator = validators.DictValidator({
        'name': validators.StringValidator(required=False),
        'spin_type': validators.ChoicesValidator(SpinType.values(),
                                                 required=False),
        'state': validators.ChoicesValidator(SpinState.values(),
                                             required=False),
        'url': validators.UrlValidator(required=False),
        'date_created': validators.DateTimeValidator(required=False)
    })
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    spin = get_or_404(Spin, id=spin_id)
    if spin.milestone != milestone:
        raise errors.NoSuchObjectError(obj_type='Spin')
    data = json_loads(request.data)
    spin_validator(data)
    if 'state' in data:
        spin.state = data['state']
    if 'name' in data:
        spin.name = data['name']
    if 'spin_type' in data:
        spin.spin_type = data['spin_type']
    if 'date_created' in data:
        spin.date_created = parse_date(data['date_created'])
    db.session.add(spin)
    db.session.commit()
    return JsonResponse(status=200)


